<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
    <title>Tensorneko Web Watcher</title>
  <script type="module" crossorigin src="/assets/index.b3ff46d5.js"></script>
  <link rel="stylesheet" href="/assets/index.6c6f085b.css">
</head>
<body>
    <noscript>
        <strong>We're sorry but <%= htmlWebpackPlugin.options.title %> doesn't work properly without JavaScript enabled.
            Please enable it to continue.</strong>
    </noscript>
    <div id="app"></div>
    
</body>
</html>
