from enum import Enum


class Colors(Enum):
    GREEN = "#C3E88D"
    YELLOW = "#FFCB6B"
    RED = "#F07178"
    BLUE = "#82AAFF"
    PURPLE = "#C792EA"
    ORANGE = "#F78C6C"
    AZURE = "#467CDA"
    SKY = "#89DDFF"
    BLACK = "#2B2B2B"
    GRAY = "#616161"
    WHITE = "#EEFFFF"
    BLANK = "#FFFFFF"


class ContinuousColors(Enum):
    RED = "#d11141",
    YELLOW = "#ffc425",
    GREEN = "#00b159",
    BLACK = "#2B2B2B",
    WHITE = "#EEFFFF"
